/*-
 * Copyright (c) 1985, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This module is believed to contain source code proprietary to AT&T.
 * Use and redistribution is subject to the Berkeley Software License
 * Agreement and your Software Agreement with AT&T (Western Electric).
 */

#ifndef lint
static char sccsid[] = "@(#)conn.c	8.1 (Berkeley) 6/6/93";
#endif /* not lint */

#include <signal.h>
#include "uucp.h"
#include <setjmp.h>
#include <ctype.h>
#include <errno.h>
#ifdef	USG
#include <termio.h>
#include <fcntl.h>
#endif
#ifndef	USG
#include <sgtty.h>
#endif
#ifdef BSD4_2
#include <sys/time.h>
#else
#include <time.h>
#endif

#define MAXC 1000

extern jmp_buf Sjbuf;
jmp_buf Cjbuf;
extern int errno, onesys;
extern const char *const sys_errlist[];
extern char MaxGrade, DefMaxGrade;

/* Parity control during login procedure */
#define	P_ZERO	0
#define	P_ONE	1
#define	P_EVEN	2
#define	P_ODD	3

#define ABORT -2

char 	*AbortOn = NULL;
char	par_tab[128];	/* must be power of two */
int	linebaudrate;	/* used for the sleep test in pk1.c */
int next_fd = -1;	/* predicted fd to close interrupted opens */

char *PCP = "PCP";	/* PC Pursuit device type */
/*
 *	catch alarm routine for "expect".
 */
void
alarmtr()
{
	signal(SIGALRM, alarmtr);
	if (next_fd >= 0) {
		if (close(next_fd))
			logent("FAIL", "ACU LINE CLOSE");
		next_fd = -1;
	}
	longjmp(Sjbuf, 1);
}

/* This template is for seismo to call ihnp4 
 * the 3 lines marked ---> will be overwritten for the appropriate city
 */
#define PCP_BAUD	3
#define PCP_PHONE	4
#define PCP_CITY	14
#define PCP_PASSWORD	16
#define PCP_RPHONE	20
#define NPCFIELDS	23

static char *PCFlds[] = {
	"PC-PURSUIT",
	"Any",
	"ACU",
	"1200",
	CNULL,
	CNULL,
	"P_ZERO",	/* Telenet insists on zero parity */
	"ABORT",
	"BUSY",		/* Abort on Busy Signal */
	CNULL,
	"\\d\\d\\r\\d\\r",	/* Get telenet's attention */
	"TERMINAL=~3-\r-TERM~3-\r-TERM~5", 	/* Terminal type ? */
	"\\r",
	"@",		/* telenet's prompt */
	"D/DCWAS/21,telenetloginstring", /* overwritten later */
	"PASSWORD",
	CNULL,		/* telenet password */
	"CONNECTED",	/* We're now talking to a Hayes in the remote city */
	"ATZ",		/* Reset it */
	"OK",
	"ATDT6907171", /* overwritten */
	"CONNECT",	
	"\\d\\r",		/* We're in !*/
	CNULL,
};

static char PCP_brand[25];
int Dcf = -1;
char *Flds[MAXC/10];
char LineType[10];
extern int LocalOnly;

/*
 *	place a telephone call to system and login, etc.
 *
 *	return codes:
 *		CF_SYSTEM: don't know system
 *		CF_TIME: wrong time to call
 *		CF_DIAL: call failed
 *		CF_NODEV: no devices available to place call
 *		CF_LOGIN: login/password dialog failed
 *
 *		>0  - file no.  -  connect ok
 */
conn(system)
char *system;
{
	int nf;
	char info[MAXC], wkpre[NAMESIZE], file[NAMESIZE];
	register FILE *fsys;
	int fcode = 0;

	nf = 0;

	fsys = fopen(SYSFILE, "r");
	if (fsys == NULL) {
		syslog(LOG_ERR, "fopen(%s) failed: %m", SYSFILE);
		cleanup(FAIL);
	}

	DEBUG(4, "finds (%s) called\n", system);
keeplooking:
	while((nf = finds(fsys, system, info, Flds)) > 0) {
		strncpy(LineType, Flds[F_LINE], 10);
		if (LocalOnly) {
			if (strcmp("TCP", LineType)
				&& strcmp("DIR", LineType)
				&& strcmp("LOCAL", LineType) ) {
					fcode = CF_TIME;
					continue;
			}
		}
		sprintf(wkpre, "%c.%.*s", CMDPRE, SYSNSIZE, Rmtname);
		if (!onesys && MaxGrade != DefMaxGrade &&
			!iswrk(file, "chk", Spool, wkpre))  {
				fcode = CF_TIME;
				continue;
		}
		/* For GTE's PC Pursuit */
		if (snccmp(LineType, PCP) == SAME) {
			FILE *dfp;
			int status;
			static struct Devices dev;

			dfp = fopen(DEVFILE, "r");
			if (dfp == NULL) {
				syslog(LOG_ERR, "fopen(%s) failed: %m",
					DEVFILE);
				cleanup(FAIL);
			}
			while ((status=rddev(dfp, &dev)) != FAIL
				&& strcmp(PCP, dev.D_type) != SAME)
					;
			fclose(dfp);
			if (status == FAIL)
				continue;
			if (mlock(PCP) == FAIL) {
				fcode = CF_NODEV;
				logent("DEVICE", "NO");
				continue;
			}
			PCFlds[PCP_BAUD] = dev.D_class;
			PCFlds[PCP_PHONE] = dev.D_calldev;
			sprintf(PCFlds[PCP_CITY], "c d/%s%s,%s",
				Flds[F_CLASS],
				index(Flds[F_CLASS], '/') == NULL ? "/12" : "",
				dev.D_arg[D_CHAT]);
			PCFlds[PCP_PASSWORD] = dev.D_line;
			strncpy(&PCFlds[PCP_RPHONE][4], Flds[F_PHONE], 7);
			strncpy(PCP_brand, dev.D_brand, sizeof(PCP_brand));
			if ((fcode = getto(PCFlds)) < 0) {
				rmlock(PCP);
				continue;
			}
			Dcf = fcode;
			fcode = login(NPCFIELDS, PCFlds, Dcf);
			if (fcode == SUCCESS)
				break;
			fcode = CF_DIAL;
			rmlock(PCP);
			/* end PC Pursuit */
		} else if ((fcode = getto(Flds)) > 0)  {
			Dcf = fcode;
			break;
		}
	}

	if (nf <= 0) {
		fclose(fsys);
		return fcode ? fcode : nf;
	}


	if (fcode >= 0) {
		DEBUG(4, "login %s\n", "called");
		setproctitle("login");
		fcode = login(nf, Flds, Dcf); }
	if (fcode < 0) {
		clsacu();
		if (fcode == ABORT) {
			fcode = CF_DIAL;
			goto  keeplooking;
		} else {
			fclose(fsys);
			return CF_LOGIN;
		}
	}
	fclose(fsys);
	fioclex(Dcf);
	return Dcf;
}

int nulldev();
int (*CU_end)() = nulldev;

/*
 *	connect to remote machine
 *
 *	return codes:
 *		>0  -  file number - ok
 *		FAIL  -  failed
 */
getto(flds)
register char *flds[];
{
	register struct condev *cd;
	int diropn();
	char *line;

	DEBUG(4, "getto: call no. %s ", flds[F_PHONE]);
	DEBUG(4, "for sys %s\n", flds[F_NAME]);

	if (snccmp(flds[F_LINE], "LOCAL") == SAME)
		line = "ACU";
	else
		line = flds[F_LINE];
#ifdef DIALINOUT
	if (snccmp(line, "ACU") != SAME)
		reenable();
#endif DIALINOUT
	CU_end = nulldev;
	if (snccmp(line, PCP) == SAME) {
		for(cd = condevs; cd->CU_meth != NULL; cd++) {
			if (snccmp(PCP_brand, cd->CU_brand) == SAME) {
				CU_end = cd->CU_clos;
				return diropn(flds);
			}
		}
		logent(PCP_brand, "UNSUPPORTED ACU TYPE");
	} else {
		for (cd = condevs; cd->CU_meth != NULL; cd++) {
			if (snccmp(cd->CU_meth, line) == SAME) {
				DEBUG(4, "Using %s to call\n", cd->CU_meth);
				return (*(cd->CU_gen))(flds);
			}
		}
		DEBUG(1, "Can't find %s, assuming DIR\n", flds[F_LINE]);
	}
	return diropn(flds);	/* search failed, so use direct */
}

/*
 *	close call unit
 *
 *	return codes:  none
 */
clsacu()
{
	/* make *sure* Dcf 